Popslide
========

Popslide is a video memory update buffer framework for the
Nintendo Entertainment System.

Background
----------
NES games need to update video memory in numerous ways during
gameplay.  But the frame rendering circuitry in the PPU has
exclusive access to video memory while it sends the picture to the
TV.  Unless the screen is completely blank, the CPU can access video
memory only during the vertical blanking interval (vblank), a short
period before one frame and the next.  So games will schedule updates
by writing them to a buffer in RAM, waiting for vblank, and then
sending the buffered updates to the PPU.

Some games keep the update buffer in a dedicated 256-byte page of
RAM.  But because the NES has only eight pages of RAM, this can use
up valuable memory that would otherwise be useful to hold the state
of game objects.  Another technique is to use part of the stack page
($0100-$01FF) for the return stack and the rest for an update buffer.
This is convenient because the NTSC vblank isn't long enough to send
more than 180 or so bytes, leaving the rest of the page for a return
stack.  Some games, such as _Battletoads_, take advantage of this
shared page by using the `PLA` instruction, which reads from the
stack with autoincrement, to load each byte from the update buffer.

Some games, such as _Thwaite_, have only one update routine because
all the updates they do are so similar.  But this constrains the
flexibility of the "shape" of each update, or how the starting
addresses, address increments, and data lengths in each type of
update relate to one another.  Others have a separate routine for
each shape, such as the seven `bgup` routines in _RHDE_.  The NES
Stripe Image format, used in _Super Mario Bros._ and other games,
stores information about the "shape" in the update buffer along
with the actual data.

Popslide is an efficient interpreter for NES Stripe Image buffer in
the stack page.  The update buffer starts at $0108 and continues
upwards, leaving room at the top for an IRQ or NMI stack frame in
case Popslide is interrupted.

Buffer format
-------------
A Stripe Image buffer consists of one or more packets followed
by a terminator.  Each packet consists of the following data:

* High byte of destination address in video memory ($00-$3F) or
  terminator ($80-$FF)
* Low byte of address
* Direction, run, and length:

    * $00-$3F: _n_ + 1 literal bytes with address increment 1
    * $40-$7F: A run of _n_ - 63 copies of the same byte with address increment 1
    * $80-$BF: _n_ - 127 literal bytes with address increment 32
    * $C0-$FF: A run of _n_ - 191 copies of the same byte with address increment 32

* 1 to 64 literal bytes of data or 1 run byte

Blitter
-------
Three versions of Popslide are planned: fully looping, partially
unrolled, and fully unrolled.  This trades off space in a game's
fixed ROM bank against time spent processing loop overhead.

Each version provides these subroutines:

* `popslide_init`  
  Sets up vectors used internally by the blitter and clears the
  update buffer.
* `popslide_clearbuf`  
  Clears the update buffer.
* `popslide_terminate_blit`  
  Terminates the buffer (by writing $FF to $0100+`popslide_used`)
  and then calls `popslide_blit`.
* `popslide_blit`  
  Sends a completed, terminated update buffer to the PPU.  Works only
  during vertical or forced blanking.  Call it with `JMP` rather than
  `JSR`, as it will `JMP popslide_return` once done.

You are expected to export this entry point:

* `popslide_return`  
  Popslide jumps here when done.  This saves cycles over `RTS`
  if there is only one caller, such as your NMI routine.
  If you want to `JSR popslide_blit`, point this at an `RTS`, as is
  done in `popslideinternal.inc` unless that's commented out.

Popslide makes available a 1-byte variable to help you keep track how
much data you have added to the update buffer.

* `popslide_buf`  
  Alias for the start of page $0100.
* `popslide_used`
  Current offset into `popslide_buf`.  Has value $08 when the buffer
  is empty.

Building a buffer
-----------------
Popslide comes with a library to build an update buffer.  It requires
a few bytes of zero page scratch space from $0000 through $0007.

* `nstripe_append`  
  Adds a stripe image with a predefined address and length to
  the transaction.  The source address is in X:A.
* `nstripe_append_yhi`  
  Adds a stripe image in a modified format that omits the high byte
  of each address and uses a low byte >= $80 as the terminator.
  This is useful for things like status bar layouts and palette
  changes.  The source address is in X:A, and the high byte of each
  destination address ($00-$3F) is in Y.  If Y >= $80, behaves as
  `nstripe_append`.
* `nstripe_append_tophi`  
  Adds a stripe image.  The source address is in X:A, and the high
  byte of destination addresses (or >= $80 for standard format) is
  in `nstripe_top`.
* `nstripe_append_src`  
  Adds a stripe image.  The source address is in `nstripe_srclo`,
  and the high byte of destination addresses (or >= $80 for standard
  format) is in `nstripe_top`.
* `nstripe_draw_rect`  
  Draws a rectangle of tile numbers.  
  `nstripe_srclo`: start of data  
  `nstripe_left`, `nstripe_top`: tile coordinates of top left corner
  `nstripe_width`, `nstripe_height`: size of data in bytes

C interface
-----------
**Caution:** Not tested.

The following symbols are also exported with a leading underscore
for use from C code compiled with cc65:

    extern unsigned char popslide_buf[];
    extern unsigned char popslide_used;
    void popslide_init();
    void popslide_clearbuf();
    void popslide_terminate_blit();
    void popslide_blit();
    void nstripe_append(const unsigned char *data);`
